/**
Code to do correlation matching.
We will try here to group best practices to do tracking using
space correlation measures (i.e. without using Fourier's transform) .

Obviously, all the code may be improved but when it is improved please
justify how, and explain what problems have been sorted in this comment.
Here are my starting design assumptions and justifications. 

In a first time I will try to use as much as possible the stl library
as it is highly reusable and well known.

A region of interest is defined generally as a set of pixels. We may provide
functions to get these set of pixel easily (i.e. rectangle, polygon, circle...)

It seems to me that the most general way to define a template is to define its shape
associated with its texture. To implement this I used a vector of relative positions that will
be used to get the value of the pixel on the image where we are looking for an "object".
The color value of the model can be stored in a different kind of container.
To let the choice of the origin to the user and to be able to check the boundaries
of the image we need to specify at the begining of the sequence
the size of the rectangle spanned by the shape
 
I just have the intuition that having this information in different structures may be 
useful but as I can not explain it feel free to propose something else.

The reason why I don't pass consistently parameters using references is merely a question
of taste: I like to construct the simple paramtres objects on the call of the function.
Check I did this only for non public function (not to confuse anybody) 

@author Manuel Boissenin
@date 14 june 2006
*/  

#include <GL/glut.h>
#include <vector>
#include <image_funcs.h>
#include <filter.h>
#include <fstream>
#include <image_fileinput.h>
#include <image_loader.h>
#include <primitives.h>
#include <mimasexception.h>
#include "corrcoef.h"
#include "correlation_tool.h"


//We use this class for speed comparison purposes. Since we are working
//with generated data the variance is null and the correlation coefficient
//is not defined for the considered data. Since we
//are not doing exactly the same operations for the classical 
//correlation coefficient calculation, speed is just an indication
template <class Arg1, class Arg2 = Arg1 >
struct special_corr: public std::binary_function< Arg1, Arg2, double >
{
  double operator()( const Arg1 &a,
                     const Arg2 &b ) {
    double sa = 0.0, sb = 0.0, saa = 0.0, sbb = 0.0, sab = 0.0;
    assert( a.size() == b.size() );
    typename Arg1::const_iterator i;
    typename Arg2::const_iterator j;
    for ( i = a.begin(), j = b.begin(); i != a.end(); i++, j++ ) {
      sa += *i;
      sb += *j;
      saa += *i * *i;
      sbb += *j * *j;
      sab += *i * *j;
    };
    int n = a.size();
    float to_do_almost_as_much_calulation = 1 /
      sqrt( ( n * saa - sa * sa + 1) *
            ( n * sbb - sb * sb + 1) );
    return ( n * sab - sa*sb );
  }
};


template <class Arg1, class Arg2 = Arg1 >
struct nb_exact_match: public std::binary_function< Arg1, Arg2, double >
{
  double operator()( const Arg1 &a,
                     const Arg2 &b ) {
    double nb_exact_match = 0;

    assert( a.size() == b.size() );
    typename Arg1::const_iterator i;
    typename Arg2::const_iterator j;
    for ( i = a.begin(), j = b.begin(); i != a.end(); i++, j++ ) {
      if (*i == *j)
	nb_exact_match++;
    };

    return nb_exact_match;
  }
};


using namespace mimas;

std::vector<point>
get_points(const image<unsigned char> &image)
{
  std::vector< point > points;
  
   for(int x = 0; x < image.getWidth(); x++)
      for(int y = 0; y < image.getHeight(); y++)
	if(image.getPixel(x,y) == 0)
	  {
	    point temp(x,y);
	    points.push_back(temp);
	  }

return points;
}


//Add a function to add automatically boundaries to the list as this 
//is prone to errors as well as to be forgotten

int main(int argc, char** argv)
{
  int retVal = 0;
  try {
    //to be able to display images
    glutInit( &argc, argv );
    x11_display disp;    
    image_mesaoutput<unsigned char > win1( &disp );
    
    image<unsigned char> current_image;
    image<unsigned char> template_image;

    //Because we are likely to have a lot of parameters it is better to put them in a file
    //the standard library make it easy to read them.

    std::string image_model_name, image_basename, format_extension, out_images_basename;
    int padding, number_of_frames, original_x_location, original_y_location, maximum_movement_px_lenght;

    std::ifstream in(argv[1]);
    in >> image_model_name >> image_basename >> padding >> format_extension >> number_of_frames
       >> original_x_location >> original_y_location >> maximum_movement_px_lenght;
    in.close();

    int current_x_location(original_x_location), current_y_location(original_y_location);

    std::ifstream image_file(image_model_name.c_str(), std::ios::binary );
    image_file >> template_image;
    std::vector<point> object_shape = get_points(template_image);

    std::cerr<<original_x_location <<", "<<original_y_location <<std::endl;

    //end of loading parameters
    //usage for the image_loader notice the hf namespace (hypothesis filter)
    hf::image_loader<unsigned char> il(current_image , image_basename, padding, format_extension, 0, number_of_frames);

	//Code for getting the image pattern to be found.  
	
    win1 << current_image << mimas::pause();

    //correlation_tool<unsigned char, special_corr> corrtool;
    //correlation_tool<unsigned char, corrcoef> corrtool;
    correlation_tool<unsigned char, nb_exact_match > corrtool;
	
    corrtool.set_pattern(template_image ,point(0,0), object_shape);

    //corrtool.set_rectangular_pattern(template_image ,point(0,0), 50, 50);

    win1 << corrtool.get_image_pattern(255)<<mimas::pause();

    //until here okay

    corrtool.set_rectangular_roi(point(original_x_location - maximum_movement_px_lenght , 
				       original_y_location - maximum_movement_px_lenght),
				 2 * maximum_movement_px_lenght + 1,
				 2 * maximum_movement_px_lenght + 1);


    while (1)
      {
	//std::vector<double> results = corrtool.correlate(current_image);
	//corrtool.display_result(results);

	point result = corrtool.get_most_similar_position(current_image);

	corrtool.set_rectangular_roi(point(result.x - maximum_movement_px_lenght,
					   result.y - maximum_movement_px_lenght),
				     2 * maximum_movement_px_lenght + 1,
				     2 * maximum_movement_px_lenght + 1);
	std::cerr<< "Position: "<< result.x << " " << result.y <<std::endl;
	il.next();
      }
  
  } catch ( mimasexception &e )
    {
      std::cerr << "An exception occurred: " <<  e.what() << std::endl;
      retVal = 1;
    }
  return retVal;
}

 

