#include "camera.h"

#ifdef HAVE_LIBMINPACK
#ifdef HAVE_LIBLAPACK
#ifdef CAMERA_CALIBRATION_ENABLE

namespace mimas {
namespace camera_calibration {

  Camera *Camera::current_cam = NULL;

  /**
   * Constructor
   */
  Camera::Camera() throw(mimasexception) {
    MMERROR(current_cam == NULL , mimasexception, ,
	     "Trying to instanciate more than one Camera class." );
    current_cam = this;
  }
  
  /**
   * Destructor
   */
  Camera::~Camera() {
    current_cam = NULL;
  }

  /**
   * Calibrate the camera.
   * 
   * This is the function one should call to get the calibration parameters.
   *
   * To calibrate the camera, one should provide four points of the chessboard 
   * in the coordinate of the chessboard (the unit is the square), and the 
   * coordinates of the same points in the picture of the chessboard taken by
   * the camera (the unit is in pixel).
   *
   * A minimum of 4 images is required to compute the calibration, but you will
   * need much more to get an accurate result.
   *
   * @param chess a vector of Chessboard classes, to know the images of the
   *        chessboard points by the camera.
   */
  void Camera::calibrate(std::vector< Chessboard > &chess) throw (mimasexception) {

    std::vector< Matrix > homographies, extrinsics;      
    Matrix intrinsics;

    chessb = chess;

    // init the vectors
    for(unsigned int i = 0; i < chess.size(); ++i) {
      homographies.push_back(Matrix(3,3));
      extrinsics.push_back(Matrix(3,4));
    }

    // find all the homographies between the chessboard and the images
    findHomographies(homographies, chess);

    // find a first guess of the camera matrix
    intrinsics = initCalibration(homographies, chess);

    // save the first approximation
    K = intrinsics;

    // find the extrinsics parameters
    findExtrinsics(extrinsics, homographies, intrinsics, chess.size());

    this->extrinsics = extrinsics;

    try {

      // maximum likelyhood estimation using Levenberg-Marquardt algorithm
      maxLikelyhood(intrinsics, chess.size());
      K_opt = intrinsics;

    } catch(mimasexception e) {

      disto = Vector(2);
      disto(0) = 0;
      disto(1) = 0;

    }

  }


  /**
   * Find the homography 
   *
   * This function finds the homography between a set of $nb$ points and $nb$ images.
   *
   * @param chess a Chessboard class
   * @param homography a 3*3 matrix, must be initialised.
   */
  void Camera::findHomography(const Chessboard &chess, Matrix &homography) {

    int nb = chess.object.size();

    Matrix system(2*nb, 9);
    Matrix Vt;

    // fill in the matrix to solve the system
    for(int i = 0; i < nb; ++i) {

      system(2*i,0) = chess.object[i].x;
      system(2*i,1) = chess.object[i].y;
      system(2*i,2) = 1;
      system(2*i,3) = 0;
      system(2*i,4) = 0;
      system(2*i,5) = 0;
      system(2*i,6) = -chess.object[i].x * chess.image[i].x;
      system(2*i,7) = -chess.object[i].y * chess.image[i].x;
      system(2*i,8) = -chess.image[i].x;

      system(2*i+1,0) = 0;
      system(2*i+1,1) = 0;
      system(2*i+1,2) = 0;
      system(2*i+1,3) = -chess.object[i].x;
      system(2*i+1,4) = -chess.object[i].y;
      system(2*i+1,5) = -1;
      system(2*i+1,6) = chess.object[i].x  * chess.image[i].y;
      system(2*i+1,7) = chess.object[i].y  * chess.image[i].y;
      system(2*i+1,8) = chess.image[i].y;

    }

    // calculate the SVD of the matrix A = U.sigma.Vt
    // we don't need to compute U or sigma.
    gesvd< double >(system, NULL, &Vt);

    // the result is the last column of the matrix V, which is the
    // last line of Vt.
    // we put the result into the homography matrix
    homography(0,0) = Vt(8,0);
    homography(0,1) = Vt(8,1);
    homography(0,2) = Vt(8,2);
    homography(1,0) = Vt(8,3);
    homography(1,1) = Vt(8,4);
    homography(1,2) = Vt(8,5);
    homography(2,0) = Vt(8,6);
    homography(2,1) = Vt(8,7);
    homography(2,2) = Vt(8,8);

  }

  /**
   * Find all the homographies
   * 
   * Given 4 points and their images for some pictures, this function
   * calculates all the homographies (one by picture).
   */
  void Camera::findHomographies(std::vector< Matrix > &homographies, std::vector< Chessboard > &chess) {

    for(unsigned i = 0; i < chess.size(); ++i) {
      findHomography(chess[i], homographies[i]);
    }

  }


  /**
   * Find a first approximation of the intrinsic matrix.
   *
   * This function returns a first approximation of the intrinsics, in the ideal
   * case where there is no noise in the pictures, and no distortion.
   */
  Camera::Matrix Camera::initCalibration(std::vector< Matrix > &homographies, std::vector< Chessboard > &chess) 
    throw (mimasexception) {

    int nbPictures = chess.size();

    Matrix system(2 * nbPictures, 6);
    Matrix U;
    Matrix Vt;

    // fill in the matrix to solve the system
    for(int i=0; i<nbPictures; ++i) {

      system(2*i,0) = homographies[i](0,0) * homographies[i](0,1);
      system(2*i,1) = homographies[i](0,0) * homographies[i](1,1)
	            + homographies[i](1,0) * homographies[i](0,1);
      system(2*i,2) = homographies[i](1,0) * homographies[i](1,1);
      system(2*i,3) = homographies[i](2,0) * homographies[i](0,1)
	            + homographies[i](0,0) * homographies[i](2,1);
      system(2*i,4) = homographies[i](2,0) * homographies[i](1,1) 
	            + homographies[i](1,0) * homographies[i](2,1);
      system(2*i,5) = homographies[i](2,0) * homographies[i](2,1);

      system(2*i+1,0) = homographies[i](0,0) * homographies[i](0,0)
	              - homographies[i](0,1) * homographies[i](0,1);
      system(2*i+1,1) = 2 * homographies[i](0,0) * homographies[i](1,0)
	              - 2 * homographies[i](0,1) * homographies[i](1,1);
      system(2*i+1,2) = homographies[i](1,0) * homographies[i](1,0)
	              - homographies[i](1,1) * homographies[i](1,1);
      system(2*i+1,3) = 2 * homographies[i](2,0) * homographies[i](0,0)
	              - 2 * homographies[i](2,1) * homographies[i](0,1);
      system(2*i+1,4) = 2 * homographies[i](2,0) * homographies[i](1,0) 
            	      - 2 * homographies[i](1,1) * homographies[i](2,1); 
      system(2*i+1,5) = homographies[i](2,0) * homographies[i](2,0)
	              - homographies[i](2,1) * homographies[i](2,1);

    }

    // calculate the SVD of the matrix A = U.sigma.Vt
    // we don't need to compute U or sigma.
    gesvd< double >(system, &U, &Vt);

    // the result is the last column of the matrix V, which is the
    // last line of Vt. We put the result into a symmetric matrix
    SymmetricMatrix sym(3,3);

    sym(0,0) = Vt(5,0);
    sym(0,1) = Vt(5,1);
    sym(0,2) = Vt(5,3);
    sym(1,1) = Vt(5,2);
    sym(1,2) = Vt(5,4);
    sym(2,2) = Vt(5,5);
    
    // the Sylvester criterion : we just check the first determinant
    // to know if the matrix is positive of negative (the matrix must
    // already be definite as a post-condition of gesvd)
    // (!!! The math needs to be checked !!!)
    if(sym(0,0) < 0) {
      sym(0,0) *= -1; 
      sym(0,1) *= -1; 
      sym(0,2) *= -1; 
      sym(1,1) *= -1; 
      sym(1,2) *= -1; 
      sym(2,2) *= -1; 
    }

    TriangularMatrix A;

    try {

      // !!! little hack to correct a bug. Would be better to find the origin
      // of the bug, but I didn't manage to.
      // Without this hack, an assertion in boost on type check is not validated.
      //disable_type_check<bool>::value = true; 

      // we need to compute the Cholesky decomposition to get A^-1
      A = mimas::pptrf< double >(sym);

      // !!! end of the hack
      //disable_type_check<bool>::value = false; 

      // inverse A^-1 to get A
      A = mimas::inv< double >(A);
      
    } catch ( mimasexception &e ) {

      MMERROR( 0, mimasexception, ,
	       "Can't compute the calibration : " << e.what() );
      
    };

    // Normalize the matrix
    A(0,0) /= A(2,2);
    A(0,1) /= A(2,2);
    A(0,2) /= A(2,2);
    A(1,1) /= A(2,2);
    A(1,2) /= A(2,2);
    A(2,2) = 1;

    return A;

  }

  // square a number
  static inline double square(double a) {
    return (a)*(a);
  }


  /**
   * Find the extrinsic parameters.
   *
   * This functions finds the extrinsic parameters (i.e. the rotation and the
   * translation) between each image of the chessboard and the chessboard itself.
   */
  void Camera::findExtrinsics(std::vector< Matrix > &extrinsics, std::vector< Matrix > &homographies,
			      Matrix &intrinsics, int nbPictures) {

    // inverse the intrinsics matrix
    Matrix A = inv< double >(intrinsics);
    Matrix temp(3,3);
    Matrix U, Vt;

    Vector h(3), r;

    double lambda;

    for(int i = 0; i < nbPictures; ++i) {
      
      // compute r1
      h(0) = homographies[i](0,0);
      h(1) = homographies[i](1,0);
      h(2) = homographies[i](2,0);

      r = boost::numeric::ublas::prod(A, h);

      // the scalar factor
      lambda = sqrt(square(r(0)) + square(r(1)) + square(r(2)));

      temp(0,0) = r(0) / lambda;
      temp(1,0) = r(1) / lambda;
      temp(2,0) = r(2) / lambda;

      // compute r2
      h(0) = homographies[i](0,1);
      h(1) = homographies[i](1,1);
      h(2) = homographies[i](2,1);

      r = boost::numeric::ublas::prod(A, h);

      temp(0,1) = r(0) / lambda;
      temp(1,1) = r(1) / lambda;
      temp(2,1) = r(2) / lambda;

      // compute r3 = r1 x r2 (cross product)
      temp(0,2) = (temp(1,0) * temp(2,1) - temp(2,0) * temp(1,1)) / lambda;
      temp(1,2) = (temp(2,0) * temp(0,1) - temp(0,0) * temp(2,1)) / lambda;
      temp(2,2) = (temp(0,0) * temp(1,1) - temp(1,0) * temp(0,1)) / lambda;

      // estimate the best rotation matrix (the closest matrix to the one 
      // we've found meeting the properties of a rotation matrix)
      gesvd< double >(temp, &U, &Vt);
      temp = boost::numeric::ublas::prod(U, Vt);

      for(int ii=0; ii<3; ++ii)
	for(int jj=0; jj<3; ++jj)
	  extrinsics[i](ii, jj) = temp(ii, jj);

      // compute t
      h(0) = homographies[i](0,2);
      h(1) = homographies[i](1,2);
      h(2) = homographies[i](2,2);

      r = boost::numeric::ublas::prod(A, h);

      extrinsics[i](0,3) = r(0) / lambda;
      extrinsics[i](1,3) = r(1) / lambda;
      extrinsics[i](2,3) = r(2) / lambda;

      // normalisation of the sign
      if(extrinsics[i](2,3) < 0) {

	for(int ii=0; ii<3; ++ii)
	  for(int jj=0; jj<4; ++jj)
	    extrinsics[i](ii,jj) *= -1;

      }

    }

  }


  /*
   * Find a first approximation of the distortion
   *
   * Important note : this function doesn't work well (crashes sometimes)
   * so it's not used in the calibration algorithm. Instead we use an
   * approximation of 0 which is just working fine
   *
  void Camera::initDistortion(void) {

    int nbPictures = chessb.size();

    Matrix D(2 * 4 * nbPictures, 2), P;
    Vector d(2 * 4 * nbPictures), M(3), R(3);

    M(2) = 1; // to be done once only

    for(int i = 0; i < nbPictures; ++i) {

      P = boost::numeric::ublas::prod(intrinsics, extrinsics[i]);

      for(int j = 0; j < 4; ++j) {

	M(0) = chessb.at(i).object.at(j).x;
	M(1) = chessb.at(i).object.at(j).y;

	R = boost::numeric::ublas::prod(P, M);

	R(0) /= R(2);
	R(1) /= R(2);

	double x2y2 = square((R(0) - intrinsics(0,2)) / intrinsics(0,0))
	            + square((R(1) - intrinsics(1,2)) / intrinsics(1,1));

	D(4*i+2*j,0) = (R(0) - intrinsics(0,2)) * x2y2;
	D(4*i+2*j,1) = (R(0) - intrinsics(0,2)) * square(x2y2);
	
	D(4*i+2*j+1,0) = (R(1) - intrinsics(1,2)) * x2y2;
	D(4*i+2*j+1,1) = (R(1) - intrinsics(1,2)) * square(x2y2);
			  
	d(4*i+2*j) = chessb.at(i).image.at(j).x - R(0);
	d(4*i+2*j+1) = chessb.at(i).image.at(j).y - R(1);

      }

    }

    Matrix DtD = boost::numeric::ublas::prod(boost::numeric::ublas::trans(D), D);
    Matrix DtDinv = inv< double >(DtD);
    Matrix DtDinvDt = boost::numeric::ublas::prod(DtDinv, boost::numeric::ublas::trans(D));
    distortion = boost::numeric::ublas::prod(DtDinvDt, d);

  }
  */

  /**
   * Calculate the projection of the chessboard points according to the given values
   * of the intrinsic matrix and the distortion coefficients, and return an array
   * containing the error between the mesured point and the calculated one.
   *
   * This function is used by the MINPACK library (in FORTRAN) to minimize the error
   * using the Levenberg-Marquardt algorithm.
   */
  void Camera::fcn(int *m, int */*n*/, double *x, double *fvec, int */*iflag*/) {

    TriangularMatrix A(3,3);
    Matrix P, extrinsic(3,3);
    Vector M(3), R(3);

    // get the distortion parameters
    double k1 = x[5], k2 = x[6];

    M(2) = 1;

    // get the intrinsics parameters
    A(0,0) = x[0];
    A(0,1) = x[1];
    A(0,2) = x[2];
    A(1,1) = x[3];
    A(1,2) = x[4];
    A(2,2) = 1;

    // for each image...
    for(int i = 0; i < *m/4; ++i){

      // get the extrinsic parameters of the current image
      for(int ii=0; ii<3; ++ii)
	for(int jj=0; jj<3; ++jj)
	  // we don't need the third column
	  extrinsic(ii, jj) = current_cam->extrinsics[i](ii, (jj == 2)?3:jj);

      // calculate the projection matrix
      P = boost::numeric::ublas::prod(A, extrinsic);

      // for each point...
      for(int j = 0; j < 4; ++j) {

	// get the object point
	M(0) = current_cam->chessb.at(i).object.at(j).x;
	M(1) = current_cam->chessb.at(i).object.at(j).y;

	// calculate the projection of the point according to the projection matrix
	R = boost::numeric::ublas::prod(P, M);

	R(0) /= R(2);
	R(1) /= R(2);
	
	// apply the distortion parameters to the calculated point
	double x2y2 = square((R(0) - A(0,1) * (R(1) - A(1,2)) / A(1,1) - A(0,2)) / A(0,0)) 
                    + square((R(1) - A(1,2)) / A(1,1));

	double K = k1 * x2y2 + k2 * square(x2y2);

	R(0) = R(0) + (R(0) - A(0,2)) * K;
	R(1) = R(1) + (R(1) - A(1,2)) * K;

	// calculate the square of the distance (norm) between the mesured point and
	// the calculated one
	fvec[4*i + j] = square(current_cam->chessb.at(i).image.at(j).x - R(0)) 
	              + square(current_cam->chessb.at(i).image.at(j).y - R(1));

      }

    }
  
  }

  /**
   * This function applies the Levenberg-Marquardt algorithm to the first
   * approximation of the intrinsic matrix.
   */
  void Camera::maxLikelyhood(Matrix &intrinsics, int nbPictures) throw (mimasexception) {

    Vector V(7);

    V(0) = intrinsics(0,0);
    V(1) = intrinsics(0,1);
    V(2) = intrinsics(0,2);
    V(3) = intrinsics(1,1);
    V(4) = intrinsics(1,2);
    V(5) = 0; // distortion(0)
    V(6) = 0; // distortion(1)

    optimisation = true;

    try {

      // call the MINPACK function
      // experimentation has shown that 20000 iteration is necessited to get
      // a result almost each time.
      V = lmdif(V, &Camera::fcn, 4*nbPictures, 20000);

    } catch(mimasexception &e) {

      optimisation = false;
      throw e;

    }

    intrinsics(0,0) = V(0);
    intrinsics(0,1) = V(1);
    intrinsics(0,2) = V(2);
    intrinsics(1,1) = V(3);
    intrinsics(1,2) = V(4);

    disto = Vector(2);
    disto(0) = V(5);
    disto(1) = V(6);

  }

}
}

#endif
#endif
#endif
